"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const Scope_1 = require("./api/Scope");
const InjectionToken_1 = require("./api/InjectionToken");
const Exception_1 = require("./Exception");
const DEFAULT_SCOPE = Scope_1.Scope.Singleton;
/*

# Composite design pattern:

         ┏━━━━━━━━━━━━━━━━━━┓
         ┃ AbstractInjector ┃
         ┗━━━━━━━━━━━━━━━━━━┛
                 ▲
                 ┃
          ┏━━━━━━┻━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━━━━━┓
          ┃                      ┃                          ┃
 ┏━━━━━━━━┻━━━━━┓   ┏━━━━━━━━━━━━┻━━━━━━━━━━━┓      ┏━━━━━━━┻━━━━━━━┓
 ┃ RootInjector ┃   ┃ AbstractCachedInjector ┃      ┃ ValueInjector ┃
 ┗━━━━━━━━━━━━━━┛   ┗━━━━━━━━━━━━━━━━━━━━━━━━┛      ┗━━━━━━━━━━━━━━━┛
                                ▲
                                ┃
                        ┏━━━━━━━┻━━━━━━━━━━━━┓
               ┏━━━━━━━━┻━━━━━━━━┓  ┏━━━━━━━━┻━━━━━━┓
               ┃ FactoryInjector ┃  ┃ ClassInjector ┃
               ┗━━━━━━━━━━━━━━━━━┛  ┗━━━━━━━━━━━━━━━┛
*/
class AbstractInjector {
    injectClass(Class, providedIn) {
        try {
            const args = this.resolveParametersToInject(Class, providedIn);
            return new Class(...args);
        }
        catch (error) {
            throw new Exception_1.Exception(`Could not inject "${Class.name}"`, error);
        }
    }
    injectFunction(fn, providedIn) {
        try {
            const args = this.resolveParametersToInject(fn, providedIn);
            return fn(...args);
        }
        catch (error) {
            throw new Exception_1.Exception(`Could not inject "${fn.name}"`, error);
        }
    }
    resolveParametersToInject(injectable, target) {
        const tokens = injectable.inject || [];
        return tokens.map(key => {
            switch (key) {
                case InjectionToken_1.TARGET_TOKEN:
                    return target;
                case InjectionToken_1.INJECTOR_TOKEN:
                    return this;
                default:
                    return this.resolve(key, injectable);
            }
        });
    }
    provideValue(token, value) {
        return new ValueInjector(this, token, value);
    }
    provideClass(token, Class, scope = DEFAULT_SCOPE) {
        return new ClassInjector(this, token, scope, Class);
    }
    provideFactory(token, factory, scope = DEFAULT_SCOPE) {
        return new FactoryInjector(this, token, scope, factory);
    }
    resolve(token, providedIn) {
        return this.resolveInternal(token, providedIn);
    }
}
class RootInjector extends AbstractInjector {
    resolveInternal(token) {
        throw new Error(`No provider found for "${token}"!.`);
    }
}
class ValueInjector extends AbstractInjector {
    constructor(parent, token, value) {
        super();
        this.parent = parent;
        this.token = token;
        this.value = value;
    }
    resolveInternal(token, target) {
        if (token === this.token) {
            return this.value;
        }
        else {
            return this.parent.resolve(token, target);
        }
    }
}
class AbstractCachedInjector extends AbstractInjector {
    constructor(parent, token, scope) {
        super();
        this.parent = parent;
        this.token = token;
        this.scope = scope;
    }
    resolveInternal(token, target) {
        if (token === this.token) {
            if (this.cached) {
                return this.cached.value;
            }
            else {
                const value = this.result(target);
                if (this.scope === Scope_1.Scope.Singleton) {
                    this.cached = { value };
                }
                return value;
            }
        }
        else {
            return this.parent.resolve(token, target);
        }
    }
}
class FactoryInjector extends AbstractCachedInjector {
    constructor(parent, token, scope, injectable) {
        super(parent, token, scope);
        this.injectable = injectable;
    }
    result(target) {
        return this.injectFunction(this.injectable, target);
    }
}
class ClassInjector extends AbstractCachedInjector {
    constructor(parent, token, scope, injectable) {
        super(parent, token, scope);
        this.injectable = injectable;
    }
    result(target) {
        return this.injectClass(this.injectable, target);
    }
}
exports.rootInjector = new RootInjector();
//# sourceMappingURL=InjectorImpl.js.map